package ru.dbotthepony.mc.otm.data.loot

import com.google.common.collect.ImmutableList
import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.functions.LootItemFunction
import net.minecraft.world.level.storage.loot.functions.LootItemFunctionType
import net.minecraft.world.level.storage.loot.parameters.LootContextParams
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.core.nbt.getJson
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.stream
import ru.dbotthepony.mc.otm.registry.MItemFunctionTypes
import java.util.Optional
import java.util.stream.Stream

class CopyTileNbtFunction(filter: Stream<out String> = Stream.empty()) : LootItemFunction, LootItemFunction.Builder {
	constructor(filter: Collection<String>) : this(filter.stream())
	constructor(vararg filter: String) : this(filter.stream())

	val filter: ImmutableList<String> = /*immutableList {
		filter.forEach(this)
	}*/ filter.collect(ImmutableList.toImmutableList())

	override fun apply(t: ItemStack, u: LootContext): ItemStack {
		val blockEntity = u.getParamOrNull(LootContextParams.BLOCK_ENTITY) ?: return t
		val result = t.tagNotNull[BlockItem.BLOCK_ENTITY_TAG] as? CompoundTag

		val data: CompoundTag

		if (blockEntity is MatteryBlockEntity) {
			data = CompoundTag()
			blockEntity.saveShared(data)
		} else {
			data = blockEntity.saveWithoutMetadata()
		}

		for (k in filter) {
			data.remove(k)
		}

		if (result == null) {
			t.tagNotNull[BlockItem.BLOCK_ENTITY_TAG] = data
		} else {
			for (k in data.allKeys) {
				result[k] = data[k]!!
			}
		}

		if ("Name" in data && "display" !in t.tagNotNull) {
			t.tagNotNull["display"] = CompoundTag().also {
				it["Name"] = Component.Serializer.toJson(Component.Serializer.fromJson(data.getJson("Name")!!)!!)
			}

			data.remove("Name")
		}

		return t
	}

	override fun getType(): LootItemFunctionType {
		return MItemFunctionTypes.COPY_TILE_NBT
	}

	override fun build(): LootItemFunction {
		return this
	}

	companion object {
		val CODEC: MapCodec<CopyTileNbtFunction> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					Codec.STRING.listOf()
						.optionalFieldOf("filter")
						.xmap({ it.orElseGet { listOf() } }, { if (it.isEmpty()) Optional.empty() else Optional.of(it) })
						.forGetter(CopyTileNbtFunction::filter),
				).apply(it, ::CopyTileNbtFunction)
			}
		}
	}
}
